from datetime import datetime
from typing import Optional, Union

from pandas import Timestamp


def to_timestamp(value: Optional[Union[str, datetime, Timestamp]]) -> Timestamp:
    """
    Convert a value to a Timestamp.

    :param value: The value to convert to a Timestamp.
    :type value: `str`, or any type convertible to `pandas.Timestamp`.
    :return: A Timestamp.
    :rtype: `pandas.Timestamp`
    """
    if value is None or value == "None" or value == "null":
        return None

    try:
        return Timestamp(value)
    except ValueError:
        raise ValueError(f"Failed to convert value {value} to Timestamp")



import logging
from datetime import datetime
from typing import Dict, List, Union

import pandas as pd
from numpy import nan

from .utils import to_timestamp

LOGGER = logging.getLogger(__name__)


def range_validity(
    values: List[Dict[str, str]],
    start_datetime: Union[str, datetime, pd.Timestamp],
    last_datetime: Union[str, datetime, pd.Timestamp],
    frequency: str,
) -> pd.DataFrame:
    """
    Include the instrument into universe by the datetime range of validity.

    :param values: The list of instrument including the symbol, valid start
        datetime and valid last datetime.
    :type values: `list[dict[str, str]]`
    :param start_datetime: The universe start datetime.
    :type start_datetime: `str`, or any type convertible by pandas `Timestamp`.
    :param last_datetime: The universe last datetime.
    :type last_datetime: `str`, or any type convertible by pandas `Timestamp`.
    :param frequency: The frequency string supported in pandas. For further
        details, please refer to
        [link](https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases)
    :type frequency: `str`
    :return: A dataframe indicating whether the instrument is included in
      the universe.
    :rtype: `pd.DataFrame`.
    """
    start_datetime = to_timestamp(start_datetime)
    last_datetime = to_timestamp(last_datetime)
    datetime_range = pd.date_range(
        start=start_datetime,
        end=last_datetime,
        freq=frequency,
        name="datetime",
    )
    default_validity = pd.Series(False, index=datetime_range)
    universe = {}
    for value in values:
        symbol = value["symbol"]
        valid_start_datetime = to_timestamp(value.get("valid_start_datetime"))
        valid_last_datetime = to_timestamp(value.get("valid_last_datetime"))
        if not valid_start_datetime:
            raise ValueError(f"Missing 'valid_start_datetime' key in value {value}")

        instrument_validity = default_validity.copy()
        if valid_start_datetime <= start_datetime:
            valid_start_datetime = start_datetime
        if valid_last_datetime is None or valid_last_datetime >= last_datetime:
            valid_last_datetime = last_datetime
        if valid_start_datetime >= valid_last_datetime:
            LOGGER.warning(
                f"No valid range is found between {start_datetime} and {last_datetime} "
                f"for {symbol}"
            )
        instrument_validity.loc[valid_start_datetime:valid_last_datetime] = True
        universe[symbol] = instrument_validity

    return pd.DataFrame(universe)


def ranking(
    values: pd.DataFrame,
    threshold_pct: float,
    tolerance_timeframes: int,
    start_datetime: Union[str, datetime, pd.Timestamp],
    last_datetime: Union[str, datetime, pd.Timestamp],
    frequency: str,
) -> pd.DataFrame:
    """
    Include the instrument into the universe by ranking.

    The instrument is selected by a threshold percentage of ranking,
    and allow to stay in the universe for a specified number of
    tolerance timeframes.

    For example, if the parameter `threshold_pct` is 0.4, and
    `tolerance_timeframes` is 21, the instrument will be selected into the
    universe only if it is ranked as the top 40% of the values,
    and will be stayed in the universe for the next 21 timeframes even
    if the condition is no longer fulfilled.

    :param values: The values are sorted in cross sectional rank. The
      columns are instruments, and the index are in datetime.
    :type values: class:`pandas.DataFrame`.
    :param threshold_pct: The threshold percentage which should be between
      0 and 1.
    :type threshold_pct: `float`.
    :param tolerance_timeframes: The number of timeframes to allow the
      instrument stays in the universe even its values is outside of the
      threshold percentage. Default is 21.
    :type tolerance_timeframes: `int`.
    :param start_datetime: The universe start datetime.
    :type start_datetime: `str`, or any type convertible by pandas `Timestamp`.
    :param last_datetime: The universe last datetime.
    :type last_datetime: `str`, or any type convertible by pandas `Timestamp`.
    :param frequency: The frequency string supported in pandas. For further
        details, please refer to
        [link](https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases)
    :type frequency: `str`
    :rtype: `pd.DataFrame`.
    """
    if not (0 <= threshold_pct <= 1):
        raise ValueError(
            f"Threshold percentage {threshold_pct} must be between 0 and 1"
        )
    datetime_range = pd.date_range(
        start=start_datetime,
        end=last_datetime,
        freq=frequency,
        name="datetime",
    )
    rank_values = values.rank(axis=1, ascending=False)
    num_instruments = values.notnull().sum(axis=1)
    threshold = num_instruments * threshold_pct
    result = (
        rank_values.le(threshold, axis=0)
        .replace(False, nan)
        .reindex(index=datetime_range)
    )

    if tolerance_timeframes > 0:
        result = result.ffill(limit=tolerance_timeframes)

    return result.fillna(False)


def rolling_validity(
    values: pd.DataFrame,
    threshold_pct: float,
    rolling_window: int,
    tolerance_timeframes: int,
    start_datetime: Union[str, datetime, pd.Timestamp],
    last_datetime: Union[str, datetime, pd.Timestamp],
    frequency: str,
) -> pd.DataFrame:
    """
    Include the instrument into the universe by rolling validity.

    The instrument is included into the universe if the values in the
    rolling timeseries exist in a specified timeframe and threshold
    percentage.

    For example, if the `threshold_pct` is 0.8 and `rolling_window` is
    21, the instrument is included into the universe only if the previous
    21 days have at least 80% valid values in the rolling basis.

    :param values: The values are sorted in cross sectional rank. The
      columns are instruments, and the index are in datetime.
    :type values: class:`pandas.DataFrame`.
    :param threshold_pct: The threshold percentage which should be between
      0 and 1.
    :type threshold_pct: `float`.
    :param rolling_window: The number of rolling timeframes that in the
      past the values exist. The number must be non-negaive.
    :type rolling_window: `int`.
    :return: A dataframe indicating whether the instrument is included in
      the universe.
    :param tolerance_timeframes: The number of timeframes to allow the
      instrument stays in the universe even its values is outside of the
      threshold percentage. Default is 21.
    :type tolerance_timeframes: `int`.
    :param start_datetime: The universe start datetime.
    :type start_datetime: `str`, or any type convertible by pandas `Timestamp`.
    :param last_datetime: The universe last datetime.
    :type last_datetime: `str`, or any type convertible by pandas `Timestamp`.
    :param frequency: The frequency string supported in pandas. For further
        details, please refer to
        [link](https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases)
    :type frequency: `str`
    :rtype: `pd.DataFrame`.
    """
    datetime_range = pd.date_range(
        start=start_datetime,
        end=last_datetime,
        freq=frequency,
        name="datetime",
    )
    result = (
        values.notnull().rolling(window=rolling_window, min_periods=1).sum()
        >= threshold_pct * rolling_window
    ).reindex(index=datetime_range)

    if tolerance_timeframes > 0:
        result = result.ffill(limit=tolerance_timeframes)

    return result



import importlib
from typing import Any, Callable, Dict, Optional

import yaml


class DelayedDataObject:
    """
    Delayed data object.

    The object is a placeholder in the JSON configuration to
    load the data from either external source or data pipeline.
    """

    _DATA_MAPPING = {}

    def __init__(self, name):
        """
        Constructor.

        Parameters
        ----------
        name : str
            Name of the data object.
        """
        self._name = name

    def __eq__(self, other) -> bool:
        """
        Equal operator.

        The object is regarded as the same if its name is the same.

        Parameters
        ----------
        other : DelayedDataObject
            Delayed data object to compare.
        """
        if not isinstance(other, DelayedDataObject):
            return False

        return self._name == other._name

    @property
    def name(self) -> str:
        """
        Return the name of the data object.
        """
        return self._name

    @property
    def values(self):
        """
        Get the values of the object.

        :return: The values of the object.
        """
        return DelayedDataObject._DATA_MAPPING[self._name]

    @classmethod
    def update_values(cls, name: str, values: Any) -> None:
        """
        Update the values in the cache.

        Parameters
        ----------
        name : str
            Name of the data object.
        values: Any
            Values of the data object.
        """
        cls._DATA_MAPPING[name] = values

    @classmethod
    def get_values(cls, name: str) -> Any:
        """
        Get the values of the object from the cache.

        Parameters
        ----------
        name : str
            Name of the data object.
        """
        return cls._DATA_MAPPING[name]


def data_representer(dumper, data):
    # fmt: off
    return dumper.represent_scalar("!data", "%s" % data)
    # fmt: on


def data_constructor(loader, node):
    name = loader.construct_scalar(node)
    return DelayedDataObject(name=name)


# fmt: off
yaml.add_constructor("!data", data_constructor)
# fmt: on


class Configuration:
    """
    Configuration.

    The class parse the configuration in a file handler or a
    text to specified data attributes, e.g. the output directory.
    """

    def __init__(self, stream, parameters: Optional[Dict[str, str]] = None):
        """
        Parameters:
        -----------
        stream: File stream.
            The file stream to read configuration from.
        parameters: Optional dictionary.
            The parameters to format in the parsed configuration.
        """
        config = yaml.load(stream, Loader=yaml.Loader)
        config = Configuration._resolve_parameters(config, parameters)
        self._config = config
        self.output_filename = Configuration._get_config(
            self._config, "output_filename"
        )
        self.intermediate_directory = Configuration._get_config(
            self._config, "intermediate_directory"
        )
        self.start_datetime = Configuration._get_config(self._config, "start_datetime")
        self.last_datetime = Configuration._get_config(self._config, "last_datetime")
        self.frequency = Configuration._get_config(self._config, "frequency")
        self.pipelines = Configuration._get_config(self._config, "pipeline")
        self.datas = Configuration._get_config(self._config, "data")

    @classmethod
    def _resolve_parameters(
        cls, items: Any, parameters: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Resolves the parameters.

        Parameters
        ----------
        items : Any
            Any types of items to resolve parameters.
        parameters : dict
            Parameters.

        Returns
        -------
        The JSON like dictionary.
        """
        if not parameters:
            return items

        if isinstance(items, str):
            try:
                return items.format(**parameters)
            except KeyError:
                raise KeyError(
                    f"Failed to format items {items} with parameters {parameters}"
                )
        elif isinstance(items, list):
            return [
                Configuration._resolve_parameters(item, parameters) for item in items
            ]
        elif isinstance(items, dict):
            return {
                Configuration._resolve_parameters(key, parameters): (
                    Configuration._resolve_parameters(item, parameters)
                )
                for key, item in items.items()
            }

        return items

    @classmethod
    def _get_config(cls, config: Dict[str, Any], key: str) -> Any:
        """
        Get a configuration with key.

        Parameters
        ----------
        config : dict
            Json like configuration object
        key: string
            Key name to look up
        """
        try:
            return config[key]
        except KeyError:
            raise KeyError(
                f"Key {key} is not found in configuration object."
                f"Available keys are: {list(config.keys())}"
            )


class DataStore:
    """
    DataStore.
    """

    def __init__(
        self,
        config: Configuration,
        custom_functions: Optional[Dict[str, Callable]] = None,
    ):
        """
        Parameters:
        -----------
        config: Configuration
            Configuration object.
        """
        self._config_datas = config.datas
        self._data_store = {
            name: DelayedDataObject(name=f"{id(self)}-{name}")
            for name in config.datas.keys()
        }
        self._custom_functions = custom_functions

    def items(self):
        """
        Iterate the items in the data store.

        Returns
        -------
        Iterator like a dictionary.
        """
        yield from self._data_store.items()

    def update_values(self, name: str, values: Any):
        """
        Update the key values.

        Parameters
        ----------
        name: string
            Name of the data object.
        values: Any
            Values of the data object.
        """
        name = f"{id(self)}-{name}"
        DelayedDataObject.update_values(name, values)

    def get(self, name: str) -> Any:
        """
        Get a data object.

        Parameters
        ----------
        name: string
            Name of the data object.
        """
        try:
            data_object = self._data_store[name]
        except KeyError:
            raise KeyError(f"Key {name} is not found in the data store")

        try:
            return data_object.values
        except KeyError:
            try:
                data_config = self._config_datas[name]
            except KeyError:
                raise KeyError(f"Key {name} is not found in configuration")
            function_name = data_config["function"]
            parameters = {}
            for param_name, parameter in data_config.get("parameters", {}).items():
                if isinstance(parameter, DelayedDataObject):
                    try:
                        parameter_values = parameter.values
                    except KeyError:
                        parameter_values = self.get(name=parameter.name)
                        self.update_values(parameter.name, parameter_values)
                elif isinstance(parameter, dict):
                    parameter_values = {
                        pm: self.get(name=pv.name)
                        if isinstance(pv, DelayedDataObject)
                        else pv
                        for pm, pv in parameter.items()
                    }
                elif isinstance(parameter, list):
                    parameter_values = [
                        self.get(name=pv.name)
                        if isinstance(pv, DelayedDataObject)
                        else pv
                        for pv in parameter
                    ]
                else:
                    parameter_values = parameter

                parameters[param_name] = parameter_values

            values = self._run_function(
                function_name=function_name, parameters=parameters
            )
            self.update_values(name, values)
            return values

    def _run_function(self, function_name: str, parameters: Dict[str, Any]) -> Any:
        """
        Run a function.

        Parameters
        ----------
        function_name: string
            Name of the function to run.
        parameters: Dict[str, Any]
            Parameters to run the function with.
        """
        data_module = importlib.import_module("fpm_universe.data")
        try:
            function = getattr(data_module, function_name)
        except AttributeError:
            try:
                function = self._custom_functions[function_name]
            except KeyError:
                raise ValueError(
                    f"Callable name {function_name} cannot be found "
                    "neither in the data module nor the customized functions"
                )

        if function is None:
            raise RuntimeError(
                "Failed to catch the exception to locate the "
                f"function {function_name}"
            )

        return function(**parameters)


class PipelineExecutor:
    def __init__(
        self,
        config: Configuration,
        custom_functions: Optional[Dict[str, Callable]] = None,
    ):
        """
        Parameters:
        -----------
        config: Configuration
            Configuration object.
        custom_functions: Optional[Dict[str, Callable]]
            Custom functions of pipelines.
        """
        self._config = config
        self._custom_functions = custom_functions

    @staticmethod
    def execute(
        data_store: DataStore,
        config: Configuration,
        pipeline: Dict[str, Any],
        custom_functions: Optional[Dict[str, Callable]] = None,
    ) -> Any:
        """
        Execute a single pipeline.

        Parameters
        ----------
        data_store: DataStore
            Data store to execute the pipeline with.
        config: Configuration
            Configuration object.
        pipeline: Dict[str, Any]
            Pipeline configuration.
        custom_functions: Optional[Dict[str, Callable]]
            Custom functions of pipelines.
        """
        name = pipeline["name"]
        function_name = pipeline["function"]
        parameters = pipeline.get("parameters", {})
        for param_name, param in parameters.copy().items():
            if not isinstance(param, DelayedDataObject):
                continue
            parameters[param_name] = data_store.get(param.name)
        data_module = importlib.import_module("fpm_universe.pipeline")
        try:
            function = getattr(data_module, function_name)
        except AttributeError:
            try:
                function = custom_functions[function_name]
            except KeyError:
                raise ValueError(
                    f"Callable name {function_name} cannot be found "
                    "neither in the pipeline module nor the customized functions"
                )

        if function is None:
            raise RuntimeError(
                "Failed to catch the exception to locate the "
                f"function {function_name}"
            )

        return name, function(
            start_datetime=config.start_datetime,
            last_datetime=config.last_datetime,
            frequency=config.frequency,
            **parameters,
        )

    def execute_all(self, data_store: DataStore) -> Dict[str, Any]:
        """
        Execute all pipelines.

        Parameters
        ----------
        data_store: DataStore
            Data store to execute the pipeline with.

        Returns
        -------
        Dict[str, Any]
            Dictionary of pipeline returns. Keys are the pipeline names
            and the values are the return values.
        """
        for pipeline in self._config.pipelines:
            yield PipelineExecutor.execute(
                config=self._config,
                pipeline=pipeline,
                data_store=data_store,
                custom_functions=self._custom_functions,
            )



import csv
import json
from enum import Enum
from functools import partial
from os import listdir
from os.path import basename
from os.path import join as fsjoin
from os.path import splitext
from typing import Any, Dict, List, Optional, Union

import jq
import pandas as pd


class FileFormat(str, Enum):
    """
    Supported file formats.
    """

    json = "json"
    csv = "csv"


class ReturnFormat(str, Enum):
    """
    Supported return formats.
    """

    default = "default"
    dict = "dict"
    dataframe = "dataframe"
    series = "series"

    def __eq__(self, __x: object) -> bool:
        if isinstance(__x, dict):
            if len(__x) != 1:
                raise TypeError(f"The dictionary {__x} length must be 1")

            return self.__eq__(list(__x.keys())[0])

        return super().__eq__(__x)


def load_all_data(
    directory: str,
    from_format: FileFormat,
    to_format: ReturnFormat,
    includes: Optional[List] = None,
):
    """
    Load all data from a directory.
    """
    file_names = listdir(directory)

    if includes:
        file_names = [
            name for name in file_names if basename(splitext(name)[0]) in includes
        ]

    if ReturnFormat.dict == to_format:
        if from_format == FileFormat.json:
            reader = json.load
        elif from_format == FileFormat.csv:
            reader = csv.DictReader
        else:
            raise ValueError(f"Unknown file format: {from_format}")
    elif ReturnFormat.dataframe == to_format:
        if from_format == FileFormat.json:
            reader = partial(
                pd.read_json,
                **(
                    {}
                    if not isinstance(to_format, dict)
                    else to_format[ReturnFormat.dataframe.value]
                ),
            )
        elif from_format == FileFormat.csv:
            reader = partial(
                pd.read_csv,
                **(
                    {}
                    if not isinstance(to_format, dict)
                    else to_format[ReturnFormat.dataframe.value]
                ),
            )
        else:
            raise ValueError(f"Unknown file format: {from_format}")
    else:
        raise ValueError(f"Unknown return format: {to_format}")

    result = {}
    for file_name in file_names:
        key_name = file_name.replace("." + from_format, "")
        path = fsjoin(directory, file_name)
        with open(path) as f:
            data = reader(f)
            result[key_name] = data

    return result


def jq_compile(
    pattern: str,
    json_filename: Optional[str] = None,
    json_input: Optional[Dict] = None,
    includes: Optional[Dict] = None,
    to_format: Optional[ReturnFormat] = ReturnFormat.default,
) -> Union[List, Dict]:
    """
    Compile a jq expression.

    Parameters
    ----------
    pattern: str
        The jq expression.
    filename: Optional[str]
        The json filename. Default is None.
    input_json: Optional[Dict]
        The input json.  Default is None.
    includes: Optional[Dict]
        The items to filter out in the result. The dictionary should
        contain the key names including in each item, while the
        values are the list of possible values in each item. For example,
        `{"symbol": ["A", "B"]}` means to filter out the result contains
        only items with symbol `A` and `B`.
    Returns
    -------
    dict
        The compiled jq expression.
    """

    def _valid(item, includes):
        return all([item[key] in values for key, values in includes.items()])

    compile = jq.compile(pattern)

    if json_filename:
        with open(json_filename) as f:
            json_input = json.load(f)

    result = compile.input(json_input).all()
    if includes:
        result = [item for item in result if _valid(item, includes)]

    if not to_format or to_format == ReturnFormat.default:
        return result

    if ReturnFormat.dict == to_format and isinstance(result, dict):
        return result
    if ReturnFormat.series == to_format and isinstance(result, list):
        to_format_parameter = to_format[ReturnFormat.series.value]
        params = {
            name: (
                [item[to_format_parameter[name]] for item in result]
                if name in to_format_parameter
                else None
            )
            for name in ["data", "index"]
        }
        params = {
            **params,
            **{
                name: to_format_parameter[name]
                for name in ["dtype", "name", "copy", "fastpath"]
                if name in to_format_parameter
            },
        }
        return pd.Series(**params)

    raise ValueError(f"Invalid return format: {to_format}. ")


def concat(
    data: Dict[str, pd.DataFrame],
    column: str,
) -> pd.DataFrame:
    """
    Concatenate a dict of dataframe into a single dataframe.

    Parameters
    ----------
    data: Dict[str, pd.DataFrame]
        The data frame.
    column: str
        The column name in the values of dataframes.
    """
    return pd.DataFrame({key: df[column] for key, df in data.items()})


def dataframe_operator(
    df: pd.DataFrame, operator: str, parameters: Dict[str, Any]
) -> Any:
    """
    Create a dataframe operator.

    Parameters
    ----------
    df: pd.DataFrame
        The dataframe.
    operator: str
        The dataframe operator name.
    parameters: Dict[str, Any]
        The parameters of the operation.

    Returns
    -------
    Any
        Any type returned from the operator function.
    """
    return getattr(df, operator)(**parameters)


def flatten(
    values: List[List[Any]],
    ascending: Optional[bool] = None,
    unique: Optional[bool] = True,
):
    """
    Flatten a list of lists.

    Parameters
    ----------
    values: List[List[Any]]
        The list of lists.
    ascending: Optional[bool]
        Indicates to sort ascending or descending. Default is None
        which means sorting is not applied.
    unique: Optional[bool]
        Indicates to return unique values. Default is True.
    """
    result = [item for value in values for item in value]
    if unique:
        result = list(set(result))
    if isinstance(ascending, bool):
        result = sorted([r for r in result if r is not None], reverse=not ascending)
    return result


def convert_str_index_to_date(df: pd.DataFrame) -> pd.DataFrame:
    """
    Convert a string index to a date index.

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe with a datetime formatted string index.

    Returns
    -------
    pd.DataFrame
        Dataframe with a pandas datetime index.
    """
    df.index = df.index.map(lambda x: pd.to_datetime(x[:10]))
    return df



import logging
from os import makedirs
from os.path import join as fsjoin

import click
import pandas as pd

from .config import Configuration, DataStore, PipelineExecutor

LOGGER = logging.getLogger(__name__)


@click.command()
@click.option("--config", "-c", help="Configuration file path.", required=True)
@click.option(
    "--parameter",
    "-p",
    multiple=True,
    help="Parameters to be formatted in the configuration.",
)
def main(config, parameter):
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    LOGGER.info("Loading parameters")
    parsed_parameters = {}
    for param in parameter:
        try:
            key, value = param.split("=")
        except ValueError:
            raise click.BadParameter(
                f"Failed to parse the parameter {parameter} into key-value "
                "pair. Please pass the parameter in the format of `{key}={value}`"
            )
        parsed_parameters[key] = value
    LOGGER.info(f"Parsed parameters: {parsed_parameters}")

    LOGGER.info("Loading configuration")
    with open(config) as fp:
        config = Configuration(stream=fp.read(), parameters=parsed_parameters)

    LOGGER.info("Loading data store")
    data_store = DataStore(config=config)

    LOGGER.info("Loading pipeline executor")
    pipeline_executor = PipelineExecutor(config=config)

    LOGGER.info("Executing the pipelines")
    pipeline_results = {}
    final_result = None
    makedirs(config.intermediate_directory, exist_ok=True)
    for name, result in pipeline_executor.execute_all(data_store=data_store):
        if not isinstance(result, pd.DataFrame):
            raise TypeError(f"Pipeline {name} does not return a DataFrame")
        pipeline_results[name] = result
        result.to_parquet(fsjoin(config.intermediate_directory, f"{name}.parquet"))
        if final_result is None:
            final_result = result
        else:
            final_result &= result

    LOGGER.info(
        f"Exporting the data to intermediate directory {config.intermediate_directory}"
    )
    for name, obj in data_store.items():
        if isinstance(obj.values, pd.DataFrame):
            path = fsjoin(config.intermediate_directory, f"{obj.name}.parquet")
            LOGGER.info(f"Exporting data {obj.name} to {path}")
            obj.values.to_parquet(path)

    LOGGER.info(
        f"Exporting the final pipeline results to output filename {config.output_filename}"
    )
    final_result.to_parquet(config.output_filename)
    LOGGER.info("Completed")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['fpm_universe']

package_data = \
{'': ['*']}

install_requires = \
['jq>=1.3.0,<2.0.0',
 'pandas>=1.3.5,<1.4.0',
 'pyarrow>=10.0.1,<11.0.0',
 'pyyaml>=6.0,<7.0']

extras_require = \
{'docs': ['Sphinx>=5.0,<6.0',
          'insipid-sphinx-theme>=0.3.6,<0.4.0',
          'myst-parser>=0.18,<0.19']}

entry_points = \
{'console_scripts': ['factor-pricing-model-universe = fpm_universe.cli:main']}

setup_kwargs = {
    'name': 'factor-pricing-model-universe',
    'version': '2022.2.0',
    'description': 'Package to build universes for factor pricing model',
    'long_description': '# Factor Pricing Model Universe\n\n<p align="center">\n  <a href="https://github.com/factorpricingmodel/factor-pricing-model-universe/actions?query=workflow%3ACI">\n    <img src="https://img.shields.io/github/workflow/status/factorpricingmodel/factor-pricing-model-universe/CI/main?label=CI&logo=github&style=flat-square" alt="CI Status" >\n  </a>\n  <a href="https://factor-pricing-model-universe.readthedocs.io">\n    <img src="https://img.shields.io/readthedocs/factor-pricing-model-universe.svg?logo=read-the-docs&logoColor=fff&style=flat-square" alt="Documentation Status">\n  </a>\n  <a href="https://codecov.io/gh/factorpricingmodel/factor-pricing-model-universe">\n    <img src="https://img.shields.io/codecov/c/github/factorpricingmodel/factor-pricing-model-universe.svg?logo=codecov&logoColor=fff&style=flat-square" alt="Test coverage percentage">\n  </a>\n</p>\n<p align="center">\n  <a href="https://python-poetry.org/">\n    <img src="https://img.shields.io/badge/packaging-poetry-299bd7?style=flat-square&logo=data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAA4AAAASCAYAAABrXO8xAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAJJSURBVHgBfZLPa1NBEMe/s7tNXoxW1KJQKaUHkXhQvHgW6UHQQ09CBS/6V3hKc/AP8CqCrUcpmop3Cx48eDB4yEECjVQrlZb80CRN8t6OM/teagVxYZi38+Yz853dJbzoMV3MM8cJUcLMSUKIE8AzQ2PieZzFxEJOHMOgMQQ+dUgSAckNXhapU/NMhDSWLs1B24A8sO1xrN4NECkcAC9ASkiIJc6k5TRiUDPhnyMMdhKc+Zx19l6SgyeW76BEONY9exVQMzKExGKwwPsCzza7KGSSWRWEQhyEaDXp6ZHEr416ygbiKYOd7TEWvvcQIeusHYMJGhTwF9y7sGnSwaWyFAiyoxzqW0PM/RjghPxF2pWReAowTEXnDh0xgcLs8l2YQmOrj3N7ByiqEoH0cARs4u78WgAVkoEDIDoOi3AkcLOHU60RIg5wC4ZuTC7FaHKQm8Hq1fQuSOBvX/sodmNJSB5geaF5CPIkUeecdMxieoRO5jz9bheL6/tXjrwCyX/UYBUcjCaWHljx1xiX6z9xEjkYAzbGVnB8pvLmyXm9ep+W8CmsSHQQY77Zx1zboxAV0w7ybMhQmfqdmmw3nEp1I0Z+FGO6M8LZdoyZnuzzBdjISicKRnpxzI9fPb+0oYXsNdyi+d3h9bm9MWYHFtPeIZfLwzmFDKy1ai3p+PDls1Llz4yyFpferxjnyjJDSEy9CaCx5m2cJPerq6Xm34eTrZt3PqxYO1XOwDYZrFlH1fWnpU38Y9HRze3lj0vOujZcXKuuXm3jP+s3KbZVra7y2EAAAAAASUVORK5CYII=" alt="Poetry">\n  </a>\n  <a href="https://github.com/ambv/black">\n    <img src="https://img.shields.io/badge/code%20style-black-000000.svg?style=flat-square" alt="black">\n  </a>\n  <a href="https://github.com/pre-commit/pre-commit">\n    <img src="https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white&style=flat-square" alt="pre-commit">\n  </a>\n</p>\n<p align="center">\n  <a href="https://pypi.org/project/factor-pricing-model-universe/">\n    <img src="https://img.shields.io/pypi/v/factor-pricing-model-universe.svg?logo=python&logoColor=fff&style=flat-square" alt="PyPI Version">\n  </a>\n  <img src="https://img.shields.io/pypi/pyversions/factor-pricing-model-universe.svg?style=flat-square&logo=python&amp;logoColor=fff" alt="Supported Python versions">\n  <img src="https://img.shields.io/pypi/l/factor-pricing-model-universe.svg?style=flat-square" alt="License">\n</p>\n\nPackage to build universes for factor pricing model. For further details, please refer\nto the [documentation](https://factor-pricing-model-universe.readthedocs.io/en/latest/)\n\n## Installation\n\nInstall this via pip (or your favourite package manager):\n\n`pip install factor-pricing-model-universe`\n\n## Usage\n\nThe library contains the pipelines to build the universe. You can\nrun the pipelines interactively in Jupyter Notebook.\n\n```python\nfrom fpm_universe import pipeline\n```\n\nAlternatively, for scheduled runs, you can create a configuration\nand run the command line entry point to create the universe.\n\n### Configuration\n\nThe configuration is in yaml format and contains a few inputs\n\n|           Name           |                                                                                            Description                                                                                            |\n| :----------------------: | :-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------: |\n|    `output_filename`     |                                                                                          Output filename                                                                                          |\n| `intermediate_directory` |                                                                       Intermediate directory to export the pipeline outputs                                                                       |\n|     `start_datetime`     |                                                                                  Start datetime of the universe                                                                                   |\n|     `last_datetime`      |                                                                                   Last datetime of the universe                                                                                   |\n|       `frequency`        | Frequency of the universe. For further details, please see the "Offset aliases" in pandas [documentation](https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases) |\n|        `pipeline`        |                                                                             List of pipelines to filter the universe                                                                              |\n|          `data`          |                                                                Defines the data used by pipeline, or referred by yaml tag `!data`                                                                 |\n\nEach pipeline returns a pandas dataframe indicating if the instrument is included into the\nuniverse on the specified date / time. For example, the pipeline returns the following\ndataframe\n\n```\n+------------+--------+-------+\n|    date    |  AAPL  | GOOGL |\n+------------+--------+-------+\n| 2022-11-17 |  True  | False |\n+------------+--------+-------+\n| 2022-11-18 |  True  |  True |\n+------------+--------+-------+\n```\n\nand it indicates AAPL is included in the universe on both 2022-11-17 and 2022-11-18\nwhile GOOGL only on 2022-11-18.\n\nBy default, the pipeline functions are imported from module `fpm_universe.pipeline`.\n\nEach data defines the method to retrieve from the source, or the operator on the\nsource data. The return type of each data is unconstrained. It can be a json-like dict,\na list, a pandas series, or even a pandas dataframe.\n\nIn the configuration, Each data can be referred by yaml tag `!data`, and it is loaded\nin lazy only when it is referred by another data object or a pipeline.\n\n### Command\n\nThe entry point `factor-pricing-model-universe` is to generate the universe regarding\nthe given configuration to the destination, with dynamically passing the parameters\nto format the configuration.\n\nThe arguments of the entry point are\n\n|        Argument        |                   Description                    |\n| :--------------------: | :----------------------------------------------: |\n|  `-c, --config TEXT`   |        Required. Configuration file path.        |\n| `-p, --parameter TEXT` | Parameters to be formatted in the configuration. |\n\nFor example, given the configuration as follows,\n\n```\noutput_filename: "{output_directory}/{date}.parquet"\nintermediate_directory: "{output_directory}/{date}"\nstart_datetime: "2015-01-01"\nlast_datetime: "{date}"\nfrequency: "B"\npipeline:\n  - name: range_validity\n    function: range_validity\n    parameters:\n      values: !data initial_validity\ndata:\n  symbols:\n    function: jq_compile\n    parameters:\n      json_filename: "{data_directory}/index/sp500/default/{date}.json"\n      pattern: "[.[] | .tickers[]] | sort | unique | .[]"\n  initial_validity:\n    function: jq_compile\n    parameters:\n      json_filename: "{data_directory}/listings/{date}.json"\n      pattern: ".[] | {{ symbol: .symbol, valid_start_datetime: .ipoDate, valid_last_datetime: .delistingDate }}"\n      includes:\n        symbol: !data symbols\n```\n\nand run the following command\n\n```\nfactor-pricing-model-universe \\\n  --config <path> \\\n  --parameter output_directory=$HOME/output \\\n  --parameter data_directory=$HOME/data \\\n  --parameter date=2022-10-20\n```\n\nthe universe dataframe is output to `$HOME/output/2022-10-20.parquet`\n(formatted with the parameter `output_directory` and `date`).\n',
    'author': 'Factor Pricing Model',
    'author_email': 'factor.pricing.model@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/factorpricingmodel/factor-pricing-model-universe',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
